@@ -136,15 +136,20 @@
   `(%define-good-modular-fun ',name ',kind ',signedp))
 
 (defmacro define-modular-fun-optimizer
-    (name ((&rest lambda-list) kind signedp &key (width (gensym "WIDTH")))
+    (name ((&rest lambda-list) kind signedp &key (width (gensym "WIDTH"))
+                                                 result-width
+                                                 (node (gensym "NODE")))
      &body body)
   (%check-modular-fun-macro-arguments name kind lambda-list)
-  (with-unique-names (call args)
+  (with-unique-names (args result-width-name)
     `(setf (gethash ',name (modular-class-funs (find-modular-class ',kind ',signedp)))
-           (lambda (,call ,width)
-             (declare (type basic-combination ,call)
-                      (type (integer 0) ,width))
-             (let ((,args (basic-combination-args ,call)))
+           (lambda (,node ,width ,(or result-width
+                                   result-width-name))
+             (declare (type basic-combination ,node)
+                      (type (integer 0) ,width)
+                      ,@(unless result-width
+                          `((ignore ,result-width-name))))
+             (let ((,args (basic-combination-args ,node)))
                (when (= (length ,args) ,(length lambda-list))
                  (destructuring-bind ,lambda-list ,args
                    (declare (type lvar ,@lambda-list))
@@ -167,7 +172,7 @@
 ;;; modular version, if it exists, or NIL. If we have changed
 ;;; anything, we need to flush old derived types, because they have
 ;;; nothing in common with the new code.
-(defun cut-to-width (lvar kind width signedp)
+(defun cut-to-width (lvar kind width signedp &optional (result-width width))
   (declare (type lvar lvar) (type (integer 0) width))
   (let ((type (specifier-type (if (zerop width)
                                   '(eql 0)
@@ -205,9 +210,9 @@
                (filter-lvar lvar
                             (if signedp
                                 (lambda (dummy)
-                                  `(mask-signed-field ,width ,dummy))
+                                  `(truly-the (signed-byte ,width) (mask-signed-field ,width ,dummy)))
                                 (lambda (dummy)
-                                  `(logand ,dummy ,(ldb (byte width 0) -1)))))
+                                  `(truly-the (unsigned-byte ,width) (logand ,dummy ,(ldb (byte width 0) -1))))))
                (do-uses (node lvar)
                  (setf (block-reoptimize (node-block node)) t)
                  (reoptimize-component (node-component node) :maybe))
@@ -274,17 +279,18 @@
                                                 (modular-fun-info
                                                  (modular-fun-info-name modular-fun))
                                                 (function
-                                                 (funcall modular-fun node width)))
+                                                 (funcall modular-fun node width result-width)))
                                               :exit-if-null)
                                         (did-something nil)
                                         (over-wide nil))
                                (unless (eql modular-fun :good)
                                  (setq did-something t
                                        over-wide t)
-                                 (change-ref-leaf
-                                  fun-ref
-                                  (find-free-fun name "in a strange place"))
-                                 (setf (combination-kind node) :full))
+                                 (unless (eq name t)
+                                   (change-ref-leaf
+                                    fun-ref
+                                    (find-free-fun name "CUT-TO-WIDTH"))
+                                   (setf (combination-kind node) :full)))
                                (unless (functionp modular-fun)
                                  (dolist (arg (basic-combination-args node))
                                    (multiple-value-bind (change wide)
@@ -296,7 +302,9 @@
                                  ;; But the outer functions don't want the type to get
                                  ;; widened and their VOPs may never be applied.
                                  (setf (node-derived-type node)
-                                       (fun-type-returns (global-ftype name)))
+                                       (fun-type-returns (global-ftype (if (eq name t)
+                                                                           fun-name
+                                                                           name))))
                                  (setf (lvar-%derived-type (node-lvar node)) nil)
                                  (ir1-optimize-combination node))
                                (values t did-something over-wide)))))))))
@@ -386,8 +394,8 @@
               ;; We cut to W not WIDTH if SIGNEDP is true, because
               ;; signed constant replacement needs to know which bit
               ;; in the field is the signed bit.
-              (let ((xact (cut-to-width x kind (if signedp w width) signedp))
-                    (yact (cut-to-width y kind (if signedp w width) signedp)))
+              (let ((xact (cut-to-width x kind (if signedp w width) signedp width))
+                    (yact (cut-to-width y kind (if signedp w width) signedp width)))
                 (declare (ignore xact yact))
                 nil) ; After fixing above, replace with T, meaning
                                         ; "don't reoptimize this (LOGAND) node any more".
@@ -428,31 +436,55 @@
                 nil) ; After fixing above, replace with T
               )))))))
 
+;;; Combine (ash (ash x 1) 1) into (ash x 2)
 (deftransform ash ((value amount))
   (let ((value-node (lvar-uses value)))
     (unless (combination-p value-node)
       (give-up-ir1-transform))
     (let ((inside-fun-name (lvar-fun-name (combination-fun value-node))))
-      (multiple-value-bind (prototype width)
-          (modular-version-info inside-fun-name :untagged nil)
-        (unless (eq (or prototype inside-fun-name) 'ash)
-          (give-up-ir1-transform))
-        (when (and width (not (constant-lvar-p amount)))
-          (give-up-ir1-transform))
-        (let ((inside-args (combination-args value-node)))
-          (unless (= (length inside-args) 2)
-            (give-up-ir1-transform))
-          (let ((inside-amount (second inside-args)))
-            (unless (and (constant-lvar-p inside-amount)
-                         (not (minusp (lvar-value inside-amount))))
-              (give-up-ir1-transform)))
-          (splice-fun-args value inside-fun-name 2)
-          (if width
+      (if (eq inside-fun-name 'ash)
+          (let* ((inside-args (combination-args value-node))
+                 (inside-amount (second inside-args))
+                 ;; Can't do anything if it shifts right erasing bits.
+                 (in-range (or (type-approximate-interval (lvar-type inside-amount))
+                               (give-up-ir1-transform)))
+                 (in-range (if (eq (interval-range-info in-range) '+)
+                               in-range
+                               (give-up-ir1-transform)))
+                 (out-range (type-approximate-interval (lvar-type amount)))
+                 (new-range (when out-range
+                              (interval-add in-range out-range))))
+
+            (when (and (or ;; Don't do it if the new amount won't shift in one direction
+                        (and new-range
+                             (not (interval-range-info new-range))
+                             (interval-range-info out-range))
+                        ;; Do not disturb the conversion to a right shift
+                        (combination-is (lvar-uses amount) '(%negate -)))
+                       ;; but not if it won't be inlined anyway
+                       (or (csubtypep (lvar-type value) (specifier-type 'word))
+                           (csubtypep (lvar-type value) (specifier-type 'sb-vm:signed-word))))
+              (give-up-ir1-transform))
+            (splice-fun-args value inside-fun-name 2)
+            `(lambda (value amount1 amount2)
+               (ash value (+ amount1 amount2))))
+          (multiple-value-bind (prototype width)
+              (modular-version-info inside-fun-name :untagged nil)
+            (unless (eq prototype 'ash)
+              (give-up-ir1-transform))
+            (when (not (constant-lvar-p amount))
+              (give-up-ir1-transform))
+            (let ((inside-args (combination-args value-node)))
+              (unless (= (length inside-args) 2)
+                (give-up-ir1-transform))
+              (let ((inside-amount (second inside-args)))
+                (unless (and (constant-lvar-p inside-amount)
+                             (not (minusp (lvar-value inside-amount))))
+                  (give-up-ir1-transform)))
+              (splice-fun-args value inside-fun-name 2)
               `(lambda (value amount1 amount2)
                  (logand (ash value (+ amount1 amount2))
-                         ,(1- (ash 1 (+ width (lvar-value amount))))))
-              `(lambda (value amount1 amount2)
-                 (ash value (+ amount1 amount2)))))))))
+                         ,(1- (ash 1 (+ width (lvar-value amount))))))))))))
 (macrolet
     ((def (left-name name kind width signedp)
        (declare (ignorable name))
@@ -463,7 +495,8 @@
             (defknown ,left-name (integer (integer 0)) (,type ,width)
                 (foldable flushable movable)
               :derive-type (make-modular-fun-type-deriver 'ash ',width ',signedp))
-            (define-modular-fun-optimizer ash ((integer count) ,kind ,signedp :width width)
+            (define-modular-fun-optimizer ash ((integer count) ,kind ,signedp :width width
+                                               :result-width result-width)
               (let ((integer-type (lvar-type integer))
                     (count-type (lvar-type count)))
                 (declare (ignorable integer-type))
@@ -481,7 +514,12 @@
                               (not (csubtypep count-type (specifier-type '(integer 0 *))))
                               (or (csubtypep integer-type (specifier-type `(unsigned-byte ,sb-vm:n-word-bits)))
                                   (csubtypep integer-type (specifier-type `(signed-byte ,sb-vm:n-word-bits)))))
-                         ',name)))))
+                         ',name)
+                        ((and (not (csubtypep integer-type (specifier-type 'word)))
+                              (not (csubtypep integer-type (specifier-type 'sb-vm:signed-word)))
+                              (csubtypep count-type (specifier-type `(integer ,(- result-width width) ,most-positive-fixnum))))
+                         (cut-to-width integer ,kind width ,signedp)
+                         t)))))
             (setf (gethash ',left-name (modular-class-versions (find-modular-class ',kind ',signedp)))
                   `(ash ,',width))
             (deftransform ,left-name ((integer count) (t (constant-arg (eql 0))))
@@ -697,3 +735,39 @@
             `(sb-vm::calc-phash val ,n-temps ,steps)
             form)))))
 )
+
+#+(or arm64 x86-64)
+(progn
+  (defknown sb-vm::truncate-mod64 (sb-vm:signed-word sb-vm:signed-word)
+      (values word sb-vm:signed-word)
+      (foldable flushable movable))
+
+  (defoptimizer (sb-vm::truncate-mod64 derive-type) ((n d) node)
+    (let ((res (truncate-derive-type-optimizer node)))
+      (when res
+        (destructuring-bind (q r) (values-type-required res)
+          (make-values-type  (list (%two-arg-derive-type q
+                                                         (specifier-type `(eql ,(ldb (byte sb-vm:n-word-bits 0) -1)))
+                                                         #'logand-derive-type-aux #'logand)
+                                   r))))))
+
+  (deftransform sb-vm::truncate-mod64 ((n d) * * :node node)
+    (let ((truncate-type (truncate-derive-type-optimizer node)))
+      (if (and truncate-type
+               (values-subtypep truncate-type
+                                (values-specifier-type `(values sb-vm:signed-word t &optional))))
+          `(multiple-value-bind (q r) (truncate n d)
+             (values (logand q ,most-positive-word)
+                     r))
+          (give-up-ir1-transform))))
+
+  (define-modular-fun-optimizer truncate
+      ((n d) :untagged nil :width width :node node)
+    (when (and (= width sb-vm:n-word-bits)
+               (not (values-subtypep (node-derived-type node)
+                                     (values-specifier-type `(values sb-vm:signed-word t &optional))))
+               (csubtypep (lvar-type n) (specifier-type 'sb-vm:signed-word))
+               (csubtypep (lvar-type d) (specifier-type 'sb-vm:signed-word)))
+      'sb-vm::truncate-mod64))
+  (setf (gethash 'sb-vm::truncate-mod64 (modular-class-versions (find-modular-class ':untagged 'nil)))
+        `(truncate ,sb-vm:n-word-bits)))
